var path = require('path');
var fs = require('fs');
var util = require("util");

var mkdirp = require('mkdirp');

exports = module.exports = function(basePath)
{
    var storage = require("../util/storage")(basePath);

    /**
     * Ensures that the content deployment root directory for Cloud CMS assets.
     *
     * This directory looks like:
     *
     *   /hosts
     *      /abc.cloudcms.net
     *          /cloudcms
     *              /<repositoryId>
     *                  /<branchId>
     *                      /<nodeId>
     *                          /<localeKey>
     *                              /content
     *
     * @param host
     * @param branchId
     * @parma nodeId
     * @param locale
     * @param callback
     *
     * @return {*}
     */
    var ensureContentDirectory = function(host, repositoryId, branchId, nodeId, locale, callback)
    {
        if (!host) {
            host = "default";
        }

        storage.ensureHostDirectory(host, function(err, hostDirectoryPath) {

            if (err) {
                callback(err);
                return;
            }

            var contentDirectoryPath = path.join(hostDirectoryPath, "content", repositoryId, branchId, nodeId, locale);

            fs.exists(contentDirectoryPath, function(exists) {

                if (!exists)
                {
                    mkdirp(contentDirectoryPath, function() {
                        callback(null, contentDirectoryPath);
                    });
                }
                else
                {
                    callback(null, contentDirectoryPath);
                }
            });
        });
    };

    /**
     * Downloads the asset from host:port/path and stores it on disk at filePath.
     *
     * @param host
     * @param port
     * @param uri
     * @param gitana driver instance
     * @param filePath
     * @param callback
     */
    var saveToDisk = function(host, port, uri, gitana, filePath, callback)
    {
        // mkdirs
        var dirPath = path.dirname(filePath);
        //var filename = path.basename(filePath);

        mkdirp(dirPath, function(err) {

            var tempFilePath = filePath + "." + new Date().getTime();
            var tempStream = fs.createWriteStream(tempFilePath);

            tempStream.on("open", function(fd) {

                // headers
                var headers = {};

                // add "authorization" for OAuth2 bearer token
                var headers2 = gitana.platform().getDriver().getHttpHeaders();
                headers["Authorization"] = headers2["Authorization"];

                var URL = process.env.GITANA_PROXY_SCHEME + "://" + process.env.GITANA_PROXY_HOST + ":" + process.env.GITANA_PROXY_PORT + uri;
                var request = require("request");
                request({
                    "method": "GET",
                    "url": URL,
                    "qs": {},
                    "headers": headers
                }).on('response', function(response) {

                    if (response.statusCode >= 200 && response.statusCode <= 204) {

                        response.pipe(tempStream).on("close", function(err) {

                            if (err) {
                                fs.unlinkSync(tempFilePath);
                            } else {

                                // remove old if exists
                                try { fs.unlinkSync(filePath); } catch (e) {}

                                // move temp file to target name
                                fs.renameSync(tempFilePath, filePath);
                            }

                            callback();

                        }).on("error", function(err) {
                            console.log("Pipe error: " + err);
                        });
                    }
                    else
                    {
                        // not found or error, make sure temp file removed from disk
                        if (fs.existsSync(tempFilePath))
                        {
                            fs.unlinkSync(tempFilePath);
                        }

                        callback({ "message": "Unable to load asset from remote store", "code": response.statusCode });
                    }

                }).end();

            });

        });
    };

    /**
     * Downloads node metadata or an attachment and saves it to disk.
     *
     * @param host
     * @param gitana driver instance
     * @param repositoryId
     * @param branchId
     * @param nodeId
     * @param attachmentId
     * @param nodePath
     * @param locale
     * @param forceReload
     * @param callback
     */
    var download = function(host, gitana, repositoryId, branchId, nodeId, attachmentId, nodePath, locale, forceReload, callback)
    {
        // base storage directory
        ensureContentDirectory(host, repositoryId, branchId, nodeId, locale, function(err, contentDirectoryPath) {

            if (err) {
                callback(err);
                return;
            }

            var filePath = contentDirectoryPath;
            if (nodePath) {
                filePath = path.join(filePath, nodePath);
            }
            if (attachmentId) {
                filePath = path.join(contentDirectoryPath, "attachments", attachmentId + path.extname(nodePath));
            } else {
                filePath = path.join(contentDirectoryPath, "metadata.json");
            }

            // if force reload, delete from disk if exist
            if (forceReload && fs.existsSync(filePath)) {
                fs.unlinkSync(filePath);
            }

            if (fs.existsSync(filePath)) {
                callback(null, filePath);
            }
            else
            {
                var assetPath = "/repositories/" + repositoryId + "/branches/" + branchId + "/nodes/" + nodeId;
                if (attachmentId) {
                    assetPath += "/attachments/" + attachmentId;
                }
                if (nodePath) {
                    assetPath += "?path=" + nodePath;
                }

                var gitanaHost = process.env.GITANA_PROXY_HOST;// || "localhost";
                var gitanaPort = process.env.GITANA_PROXY_PORT;// || 8080;

                saveToDisk(gitanaHost, gitanaPort, assetPath, gitana, filePath, function(err) {
                    if (err) {
                        console.log("ERR: " + err.message);
                        callback(err);
                    }
                    else {
                        //console.log("Fetched: " + assetPath);
                        //console.log("Retrieved from server: " + filePath);
                        callback(null, filePath);
                    }
                });
            }
        });
    };

    /**
     * Downloads a preview image for a node.
     *
     * @param host
     * @param gitana driver instance
     * @param repositoryId
     * @param branchId
     * @param nodeId
     * @param nodePath
     * @param locale
     * @param previewId
     * @param size
     * @param forceReload
     * @param callback
     */
    var preview = function(host, gitana, repositoryId, branchId, nodeId, nodePath, locale, previewId, size, forceReload, callback)
    {
        // base storage directory
        ensureContentDirectory(host, repositoryId, branchId, nodeId, locale, function(err, contentDirectoryPath) {

            if (err) {
                callback(err);
                return;
            }

            var filePath = path.join(contentDirectoryPath, "previews", previewId);

            // if force reload, delete from disk if exist
            if (forceReload && fs.existsSync(filePath)) {
                fs.unlinkSync(filePath);
            }

            if (fs.existsSync(filePath)) {
                callback(null, filePath);
            }
            else
            {
                var uri = "/repositories/" + repositoryId + "/branches/" + branchId + "/nodes/" + nodeId + "/preview/" + previewId;
                uri += "?a=1";
                if (nodePath)
                {
                    uri += "&path=" + nodePath;
                }
                if (size) {
                    uri += "&size=" + size;
                }
                if (forceReload) {
                    uri += "&force=" + forceReload;
                }

                var gitanaHost = process.env.GITANA_PROXY_HOST;// || "localhost";
                var gitanaPort = process.env.GITANA_PROXY_PORT;// || 8080;

                console.log("PREVIEW HOST: " + gitanaHost);
                console.log("PREVIEW PORT: " + gitanaPort);
                console.log("PREVIEW URL: " + uri);

                saveToDisk(gitanaHost, gitanaPort, uri, gitana, filePath, function(err) {
                    if (err) {
                        console.log("ERR: " + err.message);
                        callback(err);
                    }
                    else {
                        //console.log("Fetched: " + assetPath);
                        //console.log("Retrieved from server: " + filePath);
                        callback(null, filePath);
                    }
                });
            }
        });
    };

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    // RESULTING OBJECT
    //
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////

    var r = {};

    r.download = function(host, gitana, repositoryId, branchId, nodeId, attachmentId, nodePath, locale, forceReload, callback)
    {
        download(host, gitana, repositoryId, branchId, nodeId, attachmentId, nodePath, locale, forceReload, callback);
    };

    r.determineBranchId = function(req) {

        var branchId = "master";

        if (req.branchId) {
            branchId = req.branchId;
        }

        return branchId;
    };

    r.preview = function(host, gitana, repositoryId, branchId, nodeId, nodePath, locale, previewId, size, forceReload, callback)
    {
        preview(host, gitana, repositoryId, branchId, nodeId, nodePath, locale, previewId, size, forceReload, callback);
    };

    return r;
};

